"""
Fastlane bot config -- network
"""
__VERSION__ = "1.0.3-RESTRICTED"
__DATE__ = "02/May 2023"

from .base import ConfigBase
from . import selectors as S

import os
from dotenv import load_dotenv

load_dotenv()

from decimal import Decimal

from dotenv import load_dotenv

load_dotenv()  # take environment variables from .env.

TENDERLY_FORK = os.environ.get("TENDERLY_FORK_ID")
mp = os.environ.get("DEFAULT_MIN_PROFIT_BNT")
DEFAULT_MIN_PROFIT_BNT = Decimal('1')


class ConfigNetwork(ConfigBase):
    """
    Fastlane bot config -- network
    """
    __VERSION__ = __VERSION__
    __DATE__ = __DATE__

    # COMMONLY USED TOKEN ADDRESSES SECTION
    #######################################################################################
    ZERO_ADDRESS = "0x0000000000000000000000000000000000000000"
    USDC_ADDRESS = "0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48"
    MKR_ADDRESS = "0x9f8F72aA9304c8B593d555F12eF6589cC3A579A2"
    LINK_ADDRESS = "0x514910771AF9Ca656af840dff83E8264EcF986CA"
    WBTC_ADDRESS = "0x2260FAC5E5542a773Aa44fBCfeDf7C193bc2C599"
    ETH_ADDRESS = "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE"
    BNT_ADDRESS = "0x1F573D6Fb3F13d689FF844B4cE37794d79a7FF1C"
    WETH_ADDRESS = WETH9_ADDRESS = "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2"

    # ACCOUNTS SECTION
    #######################################################################################
    BINANCE8_WALLET_ADDRESS = "0xF977814e90dA44bFA03b6295A0616a897441aceC"
    BINANCE14_WALLET_ADDRESS = "0x28c6c06298d514db089934071355e5743bf21d60"

    # EXCHANGE IDENTIFIERS SECTION
    #######################################################################################
    BANCOR_V2_NAME = "bancor_v2"
    BANCOR_V3_NAME = "bancor_v3"
    UNISWAP_V2_NAME = "uniswap_v2"
    UNISWAP_V3_NAME = "uniswap_v3"
    SUSHISWAP_V2_NAME = "sushiswap_v2"
    CARBON_V1_NAME = "carbon_v1"
    EXCHANGE_IDS = {
        CARBON_V1_NAME: 6,
        UNISWAP_V2_NAME: 3,
        UNISWAP_V3_NAME: 4,
        BANCOR_V2_NAME: 1,
        BANCOR_V3_NAME: 2,
        SUSHISWAP_V2_NAME: 5,
    }
    UNIV2_FORKS = [UNISWAP_V2_NAME, SUSHISWAP_V2_NAME]
    SUPPORTED_EXCHANGES = list(EXCHANGE_IDS)

    # CARBON EVENTS
    #######################################################################################
    CARBON_POOL_CREATED = f"{CARBON_V1_NAME}_PoolCreated"
    CARBON_STRATEGY_CREATED = f"{CARBON_V1_NAME}_StrategyCreated"
    CARBON_STRATEGY_DELETED = f"{CARBON_V1_NAME}_StrategyDeleted"
    CARBON_STRATEGY_UPDATED = f"{CARBON_V1_NAME}_StrategyUpdated"
    CARBON_TOKENS_TRADED = f"{CARBON_V1_NAME}_TokensTraded"

    # DEFAULT VALUES SECTION
    #######################################################################################
    UNIV3_FEE_LIST = [100, 500, 3000, 10000]
    MIN_BNT_LIQUIDITY = 2_000_000_000_000_000_000
    DEFAULT_GAS = 950_000
    DEFAULT_GAS_PRICE = 0
    DEFAULT_GAS_PRICE_OFFSET = 1.09
    DEFAULT_GAS_SAFETY_OFFSET = 25_000
    DEFAULT_POLL_INTERVAL = 12
    DEFAULT_BLOCKTIME_DEVIATION = 13 * 500  # 10 block time deviation
    DEFAULT_MIN_PROFIT = DEFAULT_MIN_PROFIT_BNT
    DEFAULT_MIN_PROFIT_BNT = DEFAULT_MIN_PROFIT_BNT
    DEFAULT_MAX_SLIPPAGE = Decimal("1")  # 1%
    _PROJECT_PATH = os.path.normpath(f"{os.getcwd()}")  # TODO: FIX THIS
    DEFAULT_CURVES_DATAFILE = os.path.normpath(f"{_PROJECT_PATH}/carbon/data/curves.csv.gz")
    CARBON_STRATEGY_CHUNK_SIZE = 200
    Q96 = Decimal("2") ** Decimal("96")
    DEFAULT_TIMEOUT = 60
    CARBON_FEE = Decimal("0.002")
    BANCOR_V3_FEE = Decimal("0.0")
    DEFAULT_REWARD_PERCENT = Decimal("0.5")

    # SUNDRY SECTION
    #######################################################################################
    COINGECKO_URL = "https://tokens.coingecko.com/uniswap/all.json"

    NETWORK_ETHEREUM = S.NETWORK_ETHEREUM
    NETWORK_MAINNET = S.NETWORK_MAINNET
    NETWORK_TENDERLY = S.NETWORK_TENDERLY

    @classmethod
    def new(cls, network=None):
        """
        Return a new ConfigNetworkes object for the specified network.
        """
        if network is None:
            network = cls.NETWORK_ETHEREUM
        if network == cls.NETWORK_ETHEREUM:
            return _ConfigNetworkMainnet(_direct=False)
        elif network == cls.NETWORK_TENDERLY:
            return _ConfigNetworkTenderly(_direct=False)
        else:
            raise ValueError(f"Invalid network: {network}")

    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class _ConfigNetworkMainnet(ConfigNetwork):
    """
    Fastlane bot config -- network [Ethereum Mainnet]
    """

    NETWORK = S.NETWORK_ETHEREUM
    NETWORK_ID = "mainnet"
    NETWORK_NAME = "Ethereum Mainnet"
    DEFAULT_PROVIDER = S.PROVIDER_ALCHEMY

    # FACTORY, CONVERTER, AND CONTROLLER ADDRESSES
    #######################################################################################
    BANCOR_V3_NETWORK_INFO_ADDRESS = "0x8E303D296851B320e6a697bAcB979d13c9D6E760"
    UNISWAP_V2_FACTORY_ADDRESS = "0x5C69bEe701ef814a2B6a3EDD4B1652CB9cc5aA6f"
    UNISWAP_V2_ROUTER_ADDRESS = "0x7a250d5630B4cF539739dF2C5dAcb4c659F2488D"
    SUSHISWAP_FACTORY_ADDRESS = "0xC0AEe478e3658e2610c5F7A4A2E1777cE9e4f2Ac"
    UNISWAP_V3_FACTORY_ADDRESS = "0x1F98431c8aD98523631AE4a59f267346ea31F984"
    BANCOR_V3_POOL_COLLECTOR_ADDRESS = "0xB67d563287D12B1F41579cB687b04988Ad564C6C"
    BANCOR_V2_CONVERTER_REGISTRY_ADDRESS = "0xC0205e203F423Bcd8B2a4d6f8C8A154b0Aa60F19"
    FASTLANE_CONTRACT_ADDRESS = "0x41Eeba3355d7D6FF628B7982F3F9D055c39488cB"
    CARBON_CONTROLLER_ADDRESS = "0xC537e898CD774e2dCBa3B14Ea6f34C93d5eA45e1"
    CARBON_CONTROLLER_VOUCHER = "0x3660F04B79751e31128f6378eAC70807e38f554E"
    MULTICALL_CONTRACT_ADDRESS = "0x5BA1e12693Dc8F9c48aAD8770482f4739bEeD696"


class _ConfigNetworkTenderly(ConfigNetwork):
    """
    Fastlane bot config -- network [Ethereum Tenderly]
    """

    NETWORK = S.NETWORK_TENDERLY
    DEFAULT_PROVIDER = S.PROVIDER_TENDERLY
    NETWORK_ID = S.NETWORK_TENDERLY
    NETWORK_NAME = "tenderly"
    TENDERLY_FORK = TENDERLY_FORK

    # FACTORY, CONVERTER, AND CONTROLLER ADDRESSES
    #######################################################################################
    BANCOR_V3_NETWORK_INFO_ADDRESS = "0x8E303D296851B320e6a697bAcB979d13c9D6E760"
    UNISWAP_V2_FACTORY_ADDRESS = "0x5C69bEe701ef814a2B6a3EDD4B1652CB9cc5aA6f"
    UNISWAP_V2_ROUTER_ADDRESS = "0x7a250d5630B4cF539739dF2C5dAcb4c659F2488D"
    SUSHISWAP_FACTORY_ADDRESS = "0xC0AEe478e3658e2610c5F7A4A2E1777cE9e4f2Ac"
    UNISWAP_V3_FACTORY_ADDRESS = "0x1F98431c8aD98523631AE4a59f267346ea31F984"
    BANCOR_V3_POOL_COLLECTOR_ADDRESS = "0xB67d563287D12B1F41579cB687b04988Ad564C6C"
    BANCOR_V2_CONVERTER_REGISTRY_ADDRESS = "0xC0205e203F423Bcd8B2a4d6f8C8A154b0Aa60F19"
    FASTLANE_CONTRACT_ADDRESS = "0x41Eeba3355d7D6FF628B7982F3F9D055c39488cB"
    CARBON_CONTROLLER_ADDRESS = "0xC537e898CD774e2dCBa3B14Ea6f34C93d5eA45e1"
    CARBON_CONTROLLER_VOUCHER = "0x3660F04B79751e31128f6378eAC70807e38f554E"
    MULTICALL_CONTRACT_ADDRESS = "0x5BA1e12693Dc8F9c48aAD8770482f4739bEeD696"

    def shellcommand(self, chain_id=1):
        """
        the shell command to run to allow the bot to connect to tenderly
        """
        s = f'brownie networks delete {self.NETWORK_NAME}\n'
        s += f'brownie networks add "Ethereum" "{self.NETWORK_NAME}" '
        s += f'host=https://rpc.tenderly.co/fork/{self.TENDERLY_FORK} chainid={chain_id}'
        return s

    def __init__(self, **kwargs):
        super().__init__(**kwargs)



# coding=utf-8
"""
Contains the manager class for pools. This class is responsible for handling pools and updating their state.

(c) Copyright Bprotocol foundation 2023.
Licensed under MIT
"""
from typing import List, Dict, Any, Callable, Optional

from web3 import Web3
from web3.contract import Contract

from fastlane_bot.events.interface import Pool
from fastlane_bot.events.managers.base import BaseManager
from fastlane_bot.events.pools import pool_factory


class PoolManager(BaseManager):
    @staticmethod
    def pool_key_from_info(pool_info: Dict[str, Any]) -> str:
        """
        Get the pool key from the pool info.

        Parameters
        ----------
        pool_info : Dict[str, Any]
            The pool info.

        Returns
        -------
        str
            The pool key.

        """
        if pool_info["exchange_name"] in ["uniswap_v2", "sushiswap_v2", "uniswap_v3"]:
            return pool_info["address"]
        elif pool_info["exchange_name"] == "carbon_v1":
            return pool_info["cid"]
        elif pool_info["exchange_name"] == "bancor_v3":
            return pool_info["tkn1_address"]

    @staticmethod
    def pool_type_from_exchange_name(exchange_name: str) -> Callable:
        """
        Get the pool type from the exchange name.

        Parameters
        ----------
        exchange_name : str
            The exchange name.

        Returns
        -------
        Callable
            The pool type class.

        """
        return pool_factory.get_pool(exchange_name)

    @staticmethod
    def pool_descr_from_info(pool_info: Dict[str, Any]) -> str:
        """
        Get the pool description from the pool info.

        Parameters
        ----------
        pool_info : Dict[str, Any]
            The pool info.

        Returns
        -------
        str
            The pool description.

        """
        return (
            f"{pool_info['exchange_name']} {pool_info['pair_name']} {pool_info['fee']}"
        )

    @staticmethod
    def pool_cid_from_descr(web3: Web3, descr: str) -> str:
        """
        Get the pool CID from the description. Only used for non-Carbon pools.

        Parameters
        ----------
        web3 : Web3
            The Web3 instance.
        descr : str
            The description.

        Returns
        -------
        str
            The pool CID.

        """
        return web3.keccak(text=descr).hex()

    @property
    def pools(self) -> List[Pool]:
        """
        Get the pools from the exchanges.

        Returns
        -------
        List[Pool]
            The pools from the exchanges.
        """
        return [
            pool
            for exchange in self.exchanges.values()
            for pool in exchange.get_pools()
        ]

    def generate_pool_info(
        self,
        address,
        exchange_name,
        tkn0_address,
        tkn1_address,
        t0_symbol,
        t1_symbol,
        t0_decimals,
        t1_decimals,
        cid,
        fee,
        fee_float,
        block_number: int = None,
    ) -> Dict[str, Any]:
        """
        Generate the pool info.

        Parameters
        ----------
        address : str
            The address.
        exchange_name : str
            The exchange name.
        tkn0_address : str
            The token 0 address.
        tkn1_address : str
            The token 1 address.
        t0_symbol : str
            The token 0 symbol.
        t1_symbol : str
            The token 1 symbol.
        t0_decimals : int
            The token 0 decimals.
        t1_decimals : int
            The token 1 decimals.
        cid : str
            The cid.
        fee : Any
            The fee.
        fee_float : float
            The fee float.
        block_number : int
            The block number

        Returns
        -------
        Dict[str, Any]
            The pool info.

        """
        pool_info = {
            "last_updated_block": block_number
            if block_number is not None
            else self.web3.eth.blockNumber,
            "address": address,
            "exchange_name": exchange_name,
            "tkn0_address": tkn0_address,
            "tkn1_address": tkn1_address,
            "tkn0_symbol": t0_symbol,
            "tkn1_symbol": t1_symbol,
            "tkn0_decimals": t0_decimals,
            "tkn1_decimals": t1_decimals,
            "cid": cid,
            "tkn0_key": f"{t0_symbol}-{tkn0_address[-4:]}",
            "tkn1_key": f"{t1_symbol}-{tkn1_address[-4:]}",
            "pair_name": f"{t0_symbol}-{tkn0_address[-4:]}/{t1_symbol}-{tkn1_address[-4:]}",
            "fee_float": fee_float,
            "fee": fee,
        }
        pool_info["descr"] = self.pool_descr_from_info(pool_info)
        return pool_info

    def add_pool_info(
        self,
        address: str,
        exchange_name: str,
        fee: Any,
        fee_float: float,
        tkn0_address: str,
        tkn1_address: str,
        cid: Optional[str] = None,
        other_args: Optional[Dict[str, Any]] = None,
        contract: Optional[Contract] = None,
        block_number: int = None,
    ) -> Dict[str, Any]:
        """
        This is the main function for adding pool info.
        Parameters
        ----------
        address : str
            The address.
        exchange_name : str
            The exchange name.
        fee : Any
            The fee.
        fee_float : float
            The fee float.
        tkn0_address : str
            The token 0 address.
        tkn1_address : str
            The token 1 address.
        cid : Optional[str], optional
            The cid.
        other_args : Optional[Dict[str, Any]], optional
            The other args.
        contract : Optional[Contract], optional
            The contract.

        Returns
        -------
        Dict[str, Any]
            The pool info.


        """
        # Get or Create ERC20 contracts for each token
        t0_symbol, t0_decimals = self.get_tkn_info(tkn0_address)
        if not t0_symbol:
            return None
        t1_symbol, t1_decimals = self.get_tkn_info(tkn1_address)
        if not t1_symbol:
            return None

        # Generate pool info
        pool_info = self.generate_pool_info(
            address,
            exchange_name,
            tkn0_address,
            tkn1_address,
            t0_symbol,
            t1_symbol,
            t0_decimals,
            t1_decimals,
            cid,
            fee,
            fee_float,
            block_number,
        )

        # Add other args if necessary
        if other_args:
            pool_info.update(other_args)

        # Update cid if necessary
        if exchange_name != "carbon_v1":
            pool_info["cid"] = self.pool_cid_from_descr(self.web3, pool_info["descr"])

        # Add pool to exchange if necessary
        pool = self.get_or_init_pool(pool_info)
        assert pool, f"Pool not found in {exchange_name} pools"

        if contract:
            pool_info.update(pool.update_from_contract(contract))

        self.pool_data.append(pool_info)
        return pool_info

    def add_pool_to_exchange(self, pool_info: Dict[str, Any]):
        """
        Add a pool to the exchange.

        Parameters
        ----------
        pool_info : Dict[str, Any]
            The pool info.

        """
        pool_type = self.pool_type_from_exchange_name(pool_info["exchange_name"])
        pool = pool_type(state=pool_info)
        self.exchanges[pool_info["exchange_name"]].add_pool(pool)

    def get_pool_info(
        self,
        key: str,
        key_value: str,
        ex_name: str,
        event: Optional[Dict[str, Any]] = None,
    ) -> Optional[Dict[str, Any]]:
        """
        Get the pool info.

        Parameters
        ----------
        key : str
            The key.
        key_value : str
            The key value.
        ex_name : str
            The exchange name.
        event : Optional[Dict[str, Any]], optional
            The event, by default None.

        Returns
        -------
        Optional[Dict[str, Any]]
            The pool info.
        """
        if ex_name == "sushiswap_v2":
            ex_name = "uniswap_v2"

        if key == "address":
            key_value = self.web3.toChecksumAddress(key_value)

        return next(
            (
                self.validate_pool_info(key_value, event, pool, key)
                for pool in self.pool_data
                if pool[key] == key_value and pool["exchange_name"] == ex_name
            ),
            None,
        )

    def update_pool_data(self, pool_info: Dict[str, Any], data: Dict[str, Any]) -> None:
        """
        Update the pool data.

        Parameters
        ----------
        pool_info : Dict[str, Any]
            The pool info.
        data : Dict[str, Any]
            The data.
        """
        for pool in self.pool_data:
            if pool["cid"] == pool_info["cid"]:
                pool.update(data)
                break

    def get_or_init_pool(self, pool_info: Dict[str, Any]) -> Pool:
        """
        Get or init the pool.

        Parameters
        ----------
        pool_info : Dict[str, Any]
            The pool info.

        Returns
        -------
        Pool
            The pool.
        """
        key = self.pool_key_from_info(pool_info)
        pool = self.exchanges[pool_info["exchange_name"]].get_pool(key)
        if not pool:
            self.add_pool_to_exchange(pool_info)
            key = self.pool_key_from_info(pool_info)
            pool = self.exchanges[pool_info["exchange_name"]].get_pool(key)
        return pool



from .tradeinstruction import TradeInstruction 
from .receipthandler import TxReceiptHandler, TxReceiptHandlerBase
from .routehandler import TxRouteHandler, TxRouteHandlerBase, RouteStruct
from .submithandler import TxSubmitHandler, TxSubmitHandlerBase
from .txhelpers import TxHelpers, TxHelper
from .univ3calc import Univ3Calculator
TxHelpersBase = TxHelpers






# coding=utf-8
"""
Contains the exchange class for BancorV3. This class is responsible for handling BancorV3 events and updating the state of the pools.

(c) Copyright Bprotocol foundation 2023.
Licensed under MIT
"""
from dataclasses import dataclass
from typing import Dict, Any

from web3.contract import Contract

from .base import Pool


@dataclass
class BancorV3Pool(Pool):
    """
    Class representing a Bancor v3 pool.
    """

    exchange_name: str = "bancor_v3"

    @staticmethod
    def unique_key() -> str:
        """
        see base class.
        """
        return "tkn1_address"

    @classmethod
    def event_matches_format(cls, event: Dict[str, Any]) -> bool:
        """
        Check if an event matches the format of a Bancor v3 event.

        Parameters
        ----------
        event : Dict[str, Any]
            The event arguments.

        Returns
        -------
        bool
            True if the event matches the format of a Bancor v3 event, False otherwise.

        """
        event_args = event["args"]
        return "pool" in event_args

    def update_from_event(
        self, event_args: Dict[str, Any], data: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        See base class.
        """
        event_args = event_args["args"]
        if event_args["tkn_address"] == "0x1F573D6Fb3F13d689FF844B4cE37794d79a7FF1C":
            data["tkn0_balance"] = event_args["newLiquidity"]
        else:
            data["tkn1_balance"] = event_args["newLiquidity"]

        for key, value in data.items():
            self.state[key] = value

        data["cid"] = self.state["cid"]
        data["fee"] = self.state["fee"]
        data["fee_float"] = self.state["fee_float"]
        data["exchange_name"] = self.state["exchange_name"]
        return data

    def update_from_contract(self, contract: Contract) -> Dict[str, Any]:
        """
        See base class.
        """
        pool_balances = contract.tradingLiquidity(self.state["tkn1_address"])
        params = {
            "fee": "0.000",
            "fee_float": 0.000,
            "tkn0_balance": pool_balances[0],
            "tkn1_balance": pool_balances[1],
            "exchange_name": self.state["exchange_name"],
            "address": self.state["address"],
        }
        for key, value in params.items():
            self.state[key] = value
        return params



# coding=utf-8
"""
Contains the pool class for SushiswapV2. This class is responsible for handling SushiswapV2 pools and updating the state of the pools.

(c) Copyright Bprotocol foundation 2023.
Licensed under MIT
"""
from dataclasses import dataclass
from typing import Dict, Any

from web3.contract import Contract

from fastlane_bot.data.pools import sushiswap_v2_pools
from fastlane_bot.events.pools.base import Pool


@dataclass
class SushiswapV2Pool(Pool):
    """
    Class representing a Sushiswap pool.
    """

    exchange_name: str = "sushiswap_v2"

    @staticmethod
    def unique_key() -> str:
        """
        see base class.
        """
        return "address"

    @classmethod
    def event_matches_format(cls, event: Dict[str, Any]) -> bool:
        """
        Check if an event matches the format of a Sushiswap event.
        """
        event_args = event["args"]
        return "reserve0" in event_args and event["address"] in sushiswap_v2_pools

    def update_from_event(
        self, event_args: Dict[str, Any], data: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        See base class.
        """
        event_args = event_args["args"]
        data["tkn0_balance"] = event_args["reserve0"]
        data["tkn1_balance"] = event_args["reserve1"]
        for key, value in data.items():
            self.state[key] = value

        data["cid"] = self.state["cid"]
        data["fee"] = self.state["fee"]
        data["fee_float"] = self.state["fee_float"]
        data["exchange_name"] = self.state["exchange_name"]
        return data

    def update_from_contract(self, contract: Contract) -> Dict[str, Any]:
        """
        See base class.
        """
        reserve_balance = contract.caller.getReserves()
        params = {
            "fee": "0.003",
            "fee_float": 0.003,
            "tkn0_balance": reserve_balance[0],
            "tkn1_balance": reserve_balance[1],
            "exchange_name": "sushiswap_v2",
        }
        for key, value in params.items():
            self.state[key] = value
        return params
